#!/usr/bin/env python
# encoding: utf-8

"""Terminal UI for histdata_downloader project."""
import os
import sys
import logging
import subprocess
from datetime import date
import time

import npyscreen

from histdata_downloader.logger import log_setup
from histdata_downloader.histdata_downloader import load_available_pairs

logger = logging.getLogger(__name__)


class TestApp(npyscreen.NPSAppManaged):

    def onStart(self):
        logger.debug("On start")
        self.registerForm("MAIN", MainForm())

    def onCleanExit(self):
        logger.debug("onCleanExit called")


class MainForm(npyscreen.ActionFormV2):

    def create(self):
        logger.debug("main form method called.")

        self.type = self.add(npyscreen.TitleSelectOne, name='type',
                             max_height=2, values=['M1', 'ticks'],
                             scroll_exit=True)
        self.date_start = self.add(npyscreen.TitleDateCombo, name="Date start")
        self.date_start.value = date(2019, 1, 1)
        self.date_end = self.add(npyscreen.TitleDateCombo, name="Date end")
        self.instruments = self.add(npyscreen.TitleMultiSelect,
                                    name='instruments', max_height=5,
                                    values=load_available_pairs(),
                                    scroll_exit=True)
        self.select_all = self.add(SelectAllButton,
                                   name='select all', relx=20)
        self.unselect_all = self.add(UnselectAllButton,
                                     name='unselect all', relx=20)
        self.output_path = self.add(npyscreen.TitleFilenameCombo,
                                    name="Output path", label=True)
        self.verbosity = self.add(npyscreen.TitleSelectOne, name='verbosity',
                                  max_height=3, values=['DEBUG',
                                                        'INFO',
                                                        'WARNING'],
                                  scroll_exit=True, value=1)
        self.command = self.add(npyscreen.TitleFixedText, name="cmd",
                                editable=False,
                                value='histdata_downloader download')
        self.launch_button = self.add(LauchButton, name='Run', relx=50)
        self.log = self.add(Output, name='Output',
                            editable=True, scroll_exit=True,
                            values=['Waiting...'])

    def while_editing(self, *args):
        verb = self.selected_verbosity[0]
        cmd = "histdata_downloader -v {} download".format(verb)
        if self.type.value:
            cmd += " -t %s " % self.selected_type[0]

        if self.date_end.value:
            cmd += " -ds {} -de {}".format(self.date_start.value,
                                          self.date_end.value)
        if self.output_path.value:
            cmd += " -o {}".format(self.output_path.value)

        if self.instruments.value:
            sub_cmd = ' '.join(['-i %s' % i for i in self.selected_instruments])
            cmd += ' ' + sub_cmd

        self.command.value = cmd
        self.command.update()

    def afterEditing(self):
        self.parentApp.setNextForm(None)

    def return_as_config(self):
        logger.debug('return_as_config method called.')
        config = {'type' : self.type.values[self.type.value[0]],
                  'date_start': self.date_start.value,
                  'date_end': self.date_end.value,
                  'instruments': self.selected_instruments,
                  'output_path': self.output_path.value}
        return config

    @property
    def selected_instruments(self):
        name_field = lambda idx : self.instruments.values[idx]
        return list(map(name_field, self.instruments.value))

    @property
    def selected_type(self):
        name_field = lambda idx : self.type.values[idx]
        return list(map(name_field, self.type.value))

    @property
    def selected_verbosity(self):
        name_field = lambda idx : self.verbosity.values[idx]
        return list(map(name_field, self.verbosity.value))

def perform(cmd, log):
    with subprocess.Popen(cmd, shell=True,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE) as proc:
        for line in iter(proc.stdout.readline, b''):
            log.values.append(line.decode('ascii'))
            log.display()
        for line in iter(proc.stderr.readline, b''):
            log.values.append(line.decode('ascii'))
            log.display()


class LauchButton(npyscreen.ButtonPress):
    def whenPressed(self):
        self.parent.log.values = ['Executing %s.' % self.parent.command.value]
        self.parent.log.display()
        perform(self.parent.command.value, self.parent.log)

class SelectAllButton(npyscreen.ButtonPress):
    def whenPressed(self):
        instr = self.parent.instruments
        instr.value = [x for x in range(len(instr.values))]
        instr.display()

class UnselectAllButton(npyscreen.ButtonPress):
    def whenPressed(self):
        instr = self.parent.instruments
        instr.value = []
        instr.display


class Output(npyscreen.BoxTitle):
    _contained_widget = npyscreen.MultiLine


if __name__ == "__main__":
    App = TestApp()
    App.run()



# -*- coding: utf-8 -*-

"""Console script for histdata_downloader."""
import sys
import logging
import click

from histdata_downloader.histdata_downloader import SetDownloader
from histdata_downloader.tui import TestApp
from histdata_downloader.logger import log_setup

logger = logging.getLogger(__name__)


@click.group()
@click.option("--verbosity", "-v", default='INFO')
def main(verbosity):
    """Console script for histdata_downloader."""
    log_setup(verbosity, 'activity.log')
    return 0

@main.command()
@click.option("--type", '-t',
              default='M1',
              type=click.Choice(['M1', 'ticks'], case_sensitive=False))
@click.option('--date-start', '-ds', required=True,
              type=click.DateTime(formats=['%Y-%m-%d']))
@click.option('--date-end', '-de', required=True,
              type=click.DateTime(formats=['%Y-%m-%d']))
@click.option('--instruments', '-i', required=True, multiple=True)
@click.option('--output-path', '-o', required=True,
              type=click.Path(exists=False))
@click.option('--tqdm', default=False, is_flag=True)
def download(**kwargs):
    logger.debug('calling download command with args %s', kwargs)
    downloader = SetDownloader(kwargs)
    downloader.run()
    logger.info("Download complete !")


@main.command()
def tui():
    logger.debug("calling tui command.")
    app = TestApp()
    app.run()


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



import logging
from logging.handlers import RotatingFileHandler
import os
import re
import sys
import traceback

from datetime import datetime


def set_root_logger(loglevel=logging.INFO):
    """Setup the root logger.
    Loaded in main process."""
    rootlogger = logging.getLogger()
    rootlogger.setLevel(loglevel)
    return rootlogger

def log_setup(loglevel, logfile):
    """Initial log set up."""
    numeric_loglevel = get_loglevel(loglevel)
    root_loglevel = numeric_loglevel
    rootlogger = set_root_logger(loglevel=root_loglevel)
    log_format = logging.Formatter("%(asctime)s %(module)-15s %(funcName)-25s"
                                   "%(levelname)-8s %(message)s",
                                   datefmt="%d/%m/%Y %H:%M:%S")
    f_handler = file_handler(numeric_loglevel, logfile, log_format)
    s_handler = stream_handler(numeric_loglevel)

    rootlogger.addHandler(f_handler)
    rootlogger.addHandler(s_handler)
    disable_existing_logger()
    logging.info("Log level set to: %s", loglevel.upper())

def get_loglevel(loglevel):
    """ Check valid log level supplied and return numeric log level """
    numeric_level = getattr(logging, loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError("Invalid log level: %s" % loglevel)

    return numeric_level


def file_handler(loglevel, logfile, log_format):
    """ Add a logging rotating file handler """
    log_file = RotatingFileHandler(logfile, 'a', 10e6, backupCount=1)
    log_file.setFormatter(log_format)
    log_file.setLevel(loglevel)
    return log_file


def stream_handler(loglevel):
    """ Add a logging cli handler """
    log_format = logging.Formatter("%(asctime)s %(levelname)-8s %(message)s",
                                   datefmt="%d/%m/%Y %H:%M:%S")
    log_console = logging.StreamHandler(sys.stdout)
    log_console.setFormatter(log_format)
    log_console.setLevel(loglevel)
    return log_console

def disable_existing_logger():
    """disable_existing_logger"""
    logger_to_disable = ['urllib3', 'urllib3.connectionpool', 'numexpr.utils']
    for _ in logger_to_disable:
        logging.getLogger(_).setLevel(logging.CRITICAL)



# -*- coding: future_fstrings -*-

"""Main module."""
import os
import sys
from datetime import datetime, timedelta
import logging
import re

import bs4 as bs
import h5py
import pandas as pd
import requests
from tqdm import tqdm
import zipfile

logger = logging.getLogger(__name__)

HISTDATA_URL = 'https://www.histdata.com/'


def date_formatter(raw_string):
    """Return a well formatted date as string."""
    date = raw_string[:8]
    date = f'{date[:4]}-{date[4:6]}-{date[6:8]}'
    time = raw_string[9:]
    if len(raw_string) > 15:
        time = f'{time[0:2]}:{time[2:4]}:{time[4:6]}.{time[6:]}'
    else:
        time = f'{time[0:2]}:{time[2:4]}:{time[4:6]}'
    return ' '.join((date, time))


class DataSet:
    """TickSet object.

    Each set to download from histdata's website is represented as an
    object. The set can simply be downloaded by calling the "get" method.

    :args::
        instrument (str): instrument that you want to get the data example
                          "EURUSD"
        year (int) : the year that you want to get the data e.g. 2019
        month (int) : the month that you want to get the data e.g. 6

    :example::
        >>> tick_set = TickSet(EURUSD', 2019, 1, type='M1') #Jan 2019, 1 minute
        >>> tick_set.get()
    """
    HISTDATA_URL = HISTDATA_URL
    data_url = f'{HISTDATA_URL}download-free-forex-historical-data/?/ascii'
    ticks_url = f'{data_url}/tick-data-quotes'
    ohlc_url = f'{data_url}/1-minute-bar-quotes'
    post_url = f'{HISTDATA_URL}get.php'

    def __init__(self, instrument, year, month, type='M1', tmp_dir="/tmp"):
        logger.debug("tick set object init with args : %s | %i | %i",
                     instrument, year, month)
        self._type = None
        self._tmp_dir = tmp_dir
        self._zip_name = None
        self._csv_path = None
        self.instrument = instrument
        self.year = year
        self.month = month
        self.type = type

    @property
    def url(self):
        """Return url dynamically using provided informations."""
        if self.type.upper() == 'M1':
            data_url = self.ohlc_url
        elif self.type.lower() == 'ticks':
            data_url = self.ticks_url
        return f'{data_url}/{self.instrument}/{self.year}/{self.month}'

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, type):
        if type.lower() in ['m1', 'ticks'] or type.upper() in ['M1', 'TICKS']:
            self._type = type
        else:
            raise ValueError('Type must be M1 or ticks')

    @staticmethod
    def _header(referer=None):
        header = {"User-Agent": "Mozilla/5.0",
                  "Accept-Encoding": None,
                  "Referer": referer}
        return header

    def __str__(self):
        return f"{self.instrument}/y{self.year}/m{self.month}"

    def _download(self, session=None, tqdm_support=True):
        logger.debug("Start download.")
        if not session:
            session = requests.Session()
        response = session.get(self.url)
        form_data = self._get_form_to_send(response)
        response = session.post(self.post_url, form_data, stream=True,
                                headers=self._header(referer=self.url))
        content_disposition = response.headers.get("content-disposition")
        content_length = int(response.headers.get("content-length"))
        self._zip_name = str(content_disposition).split("=")[-1]

        with open(os.path.join(self._tmp_dir, self._zip_name), 'wb') as z_file:
            chunk_size = 1024
            for data in tqdm(response.iter_content(chunk_size=chunk_size),
                             total= content_length // chunk_size,
                             desc='Download set %s' % self.__str__(),
                             ascii=True if not tqdm_support else False,
                             file=sys.stdout):
                z_file.write(data)

        logger.debug("Download complete")
        session.close()

    def _get_form_to_send(self, response):
        soup = bs.BeautifulSoup(response.content, 'lxml')
        form = soup.find_all('form')[0]
        data_to_send = dict()
        for field in form.find_all('input'):
            data_to_send[field.attrs['name']] = field.attrs['value']
        return data_to_send

    def _extract(self, purge=True):
        """This will extract the files from the zip archive."""
        logger.debug("Extracting set : %s", self.__str__())
        path = os.path.join(self._tmp_dir, self._zip_name)
        with zipfile.ZipFile(path, 'r') as zip_r:
            output_path = str(path).split('.')[0]
            zip_r.extractall(output_path)
        csv_name = [x for x in os.listdir(output_path) if x.endswith(".csv")]
        assert len(csv_name) == 1
        self._csv_path = os.path.join(output_path, csv_name[0])
        if purge:
            logger.debug("Removing %s", self._zip_name)
            os.remove(os.path.join(self._tmp_dir, self._zip_name))
        logger.debug("Extraction complete")

    def _load_csv(self):
        """Load data from the CSV and return it as a pandas.DataFrame."""
        logger.debug("Loading csv.")
        if self.type.upper() == 'M1':
            names = ['timestamp', 'open', 'high', 'low', 'close', 'vol']
        elif self.type.lower() == 'ticks':
            names = ["timestamp", "bid", "ask", "vol"]
        else:
            raise ValueError('Invalid type data')
        df = pd.read_csv(self._csv_path,
                         names=names,
                         sep=';' if self.type.upper() == 'M1' else ',')
        df['timestamp'] = df['timestamp'].apply(date_formatter)
        df['timestamp'] = pd.to_datetime(df['timestamp'])
        return df

    def _save(self, path_hdf):
        """This will save the dataframe in a hdf5 file."""
        logger.debug("Saving data in hdf5 file.")
        df = self._load_csv()
        df.to_hdf(path_hdf, self.__str__(), table=True, mode="a")
        path = "/".join(self._csv_path.split("/")[:-1])

    def get(self, session=None, output_path=None, tqdm_support=False):
        """Get method. Run this for process a set."""
        logger.debug("get set : %s ", self.__str__())
        self._download(session, tqdm_support=tqdm_support)
        self._extract()
        if not output_path:
            raise Exception('No output_path')
        self._save(output_path)

    @property
    def should_update(self):
        today = datetime.today()
        date_set = datetime(year=self.year, month=self.month, day=1)
        if (today - date_set) > timedelta(days=30):
            return False
        return True

class SetDownloader:
    HISTDATA_URL = HISTDATA_URL

    def __init__(self, config):
        logger.debug("SetDownloader called with args: %s", config)
        self.session = self._init_session()
        self.instruments = config.get('instruments')
        self.date_start = config.get('date_start')
        self.date_end = config.get('date_end')
        self.output_path = config.get('output_path')
        self.type = config.get('type')
        self.tqdm_support = config.get('tqdm')

    def _init_session(self):
        session = requests.Session()
        try:
            response = session.get(self.HISTDATA_URL)
            if response.status_code != 200:
                raise ConnectionError('HTTP error %i', response.status_code)
        except requests.exceptions.RequestsException as e:
            logger.exception(e)
            raise ConnectionError("Histdata's website is unreachable. "
                                  "Check your internet connection.")
        return session

    @property
    def all_sets(self):
        for instrument in self.instruments:
            for date in pd.date_range(self.date_start, self.date_end,
                                      freq='M', closed='left'):
                yield DataSet(instrument, date.year, date.month,
                              type=self.type)

    def run(self):
        logger.debug('Run method called.')
        try:
            index = self._get_index()
        except OSError:
            logger.info('No index')
            index = []
        for dataset in self.all_sets:
            if str(dataset) in index:
                logger.info('Skipping %s, already in dataset.', str(dataset))
            else:
                logger.info("Processing set : %s", str(dataset))
                dataset.get(session=self.session, output_path=self.output_path,
                            tqdm_support=self.tqdm_support)
                logger.info("Done. Set : %s [saved]", dataset)
        self.session.close()

    def _get_index(self):
        logger.debug("Get index")
        index = []
        hdf5_file = h5py.File(self.output_path, 'r')
        for instr in self._get_available_instruments(hdf5_file):
            for year in self._get_available_years_for_instr(hdf5_file, instr):
                for month in self._get_available_months(hdf5_file, instr, year):
                    index.append(instr + '/' + year + '/' + month)
        hdf5_file.close()
        logger.debug('Index : %s ', index)
        return index

    def _get_available_instruments(self, hdf5_file):
        return list(hdf5_file.keys())

    def _get_available_years_for_instr(self, hdf5_file, instrument):
        return list(hdf5_file[instrument].keys())

    def _get_available_months(self, hdf5_file, instrument, year):
        return list(hdf5_file[instrument + '/' + year])


def load_available_pairs(url='https://www.histdata.com/download-free-forex-data/?/ascii/1-minute-bar-quotes'):
    logger.debug("Loading available pairs from histdata")
    response = requests.get(url)
    soup = bs.BeautifulSoup(response.content, 'html.parser')
    table = soup.find_all('table')[0]
    instr = table.find_all('strong')
    instr = list(map(str, instr))
    regex = re.compile(r'[A-Z]{3}/[A-Z]{3}')
    parser = lambda x : ''.join(regex.findall(x)[0].split('/'))
    return list(map(parser, instr))
